@@ -28,6 +28,21 @@
 char *gets(char *);
 #endif
 
+// Extra pthread_attr_t field:
+#define _a_transferredcanvases __u.__s[9]
+
+int emscripten_pthread_attr_gettransferredcanvases(const pthread_attr_t *a, const char **str)
+{
+	*str = (const char *)a->_a_transferredcanvases;
+	return 0;
+}
+
+int emscripten_pthread_attr_settransferredcanvases(pthread_attr_t *a, const char *str)
+{
+	a->_a_transferredcanvases = (int)str;
+	return 0;
+}
+
 int _pthread_getcanceltype()
 {
 	return pthread_self()->cancelasync;
